// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Module: RatioEeCorrector
// Author: Hassan Siddiqui
//
// Description: 
//
// Ratio source implementation of an Encircled Energy Corrector class
//
// Modifications: See ChangeLog
// 
// $Id: RatioEeCorrector.cc,v 1.2 2012/06/08 07:26:54 rsaxton Exp $

/* General */

/* SAS-specific */
#include "Tracer.h"

/* Local */
#include "NumericalRange.h"
#include "RatioEeCorrector.h"
#include "PixelIterator.h"
#include "ArfDataServer.h"
#include "CalServer.h"
#include "Region.h"

/* Constructor */
RatioEeCorrector::RatioEeCorrector(Xmm::Mode windowMode, const GeoRegion &region)
{
  PROC;

   /* Need to find the Y extent of the spatial region. This is
      difficult. Here we simplify by assuming that the dss only
      contains one box shape */
     double rawYextent=0.0;
     for (unsigned j=0; j<region[0].size(); j++) {
                                                                                
       if (region[0][0]->type() == ShapeBasics::BOX ||
              region[0][0]->type() == ShapeBasics::ROTBOX) 
          {
               rawYextent = region[0][j]->radii()[1]; 
          }
     }

     if (rawYextent<1.0) 
        errstr<< "Burst mode spatial region doesn't contain a BOX"
        << Sas::error("InvalidSpatialRegion");

   /* Calculate the fraction of the full burst region which has 
      been extracted */
     if (windowMode == Xmm::FAST_BURST) {
        _ratio = rawYextent / (180.0 * 20 * 4.1);
     } else if (windowMode == Xmm::FAST_TIMING) {
        _ratio = rawYextent / (200.0 * 20 * 4.1);
     } else {
        _ratio = 1.0;
     }

     _ratio = _ratio > 1.0 ? 1.0:_ratio;   // Max allowed = 1.0 

     errstr << "RatioEeCorrector: ratio="<<_ratio
	    << " rawYextent="<<rawYextent<< endl
            << Sas::message(Msg::App, Msg::Verbose); 

  ENDPROC;
}

/* Destructor */
RatioEeCorrector::~RatioEeCorrector()
{
  PROC;
  ENDPROC;
}

vector<double> RatioEeCorrector::Fraction( const vector<double>& energyvector ) const
{
  PROC;

  errstr << "RatioEeCorrector: Fraction of psf enclosed " 
	 << _ratio << endl
	 << Sas::message(Msg::App, Msg::Verbose); 

   vector<double> factors;
                                                                                
  for (unsigned i = 0; i < energyvector.size(); ++i)
    {
     factors.push_back(_ratio);
    }
                                                                                
  RETURN factors;
}
